"""A module for baseline classifiers.

"""

import warnings
from typing import Union, List
from typeguard import typechecked
from tqdm import tqdm
import pyspark
from pyspark.ml import Pipeline
from pyspark.sql import functions as F
import pyspark.ml.classification as spark_cl
import mlflow

import pyspark_ds_toolbox.ml.classification.eval as cl_eval
from pyspark_ds_toolbox.ml.data_prep.features_vector import get_features_vector 
from pyspark_ds_toolbox.ml.data_prep.class_weights import binary_classifier_weights
from pyspark_ds_toolbox.ml.feature_importance.native_spark import extract_features_score

@typechecked
def baseline_binary_classfiers(
    dfs: pyspark.sql.dataframe.DataFrame,
    dfs_test: pyspark.sql.dataframe.DataFrame,
    id_col: str,
    target_col: str,
    num_features: Union[List[str], None] = None,
    cat_features: Union[List[str], None] = None,
    weight_on_target: bool = False,
    log_mlflow_run: bool = False,
    mlflow_experiment_name: Union[None, str] = None,
    artifact_stage_path: Union[None, str] = None
) -> dict:
    """Function that fit models that could be used as baseline model.

    This function will:
     1) Add a features vecto to dfs and dfs_test (see pyspark_ds_toolbox.ml.classification.data_prep.get_features_vector());
     2) Fit the following models LogisticRegression, DecisionTreeClassifier, RandomForestClassifier and GBTClassifier, without any tunning;
     3) Extracts from the trained models the features score (see pyspark_ds_toolbox.ml.feature_importance.native_spark.extract_features_score);
     4) Use the fitted models to predict on the dfs_test;
     5) Compute evaluation metrics on the test data (see pyspark_ds_toolbox.ml.classification.eval module)

    Args:
        dfs (pyspark.sql.dataframe.DataFrame): A training DataFrameSpark.
        dfs_test (pyspark.sql.dataframe.DataFrame): A test DataFrameSpark. Metrics reported are computed with the prediction values from this data.
        id_col (str): Column name of the id. Used to compute the confusion matrix.
        target_col (str): Target to be predicted. Must be of values 1 and 0.
        num_features (Union[List[str], None], optional): List of the columns names of numerical features. Defaults to None.
        cat_features (Union[List[str], None], optional): List of the columns names of the categorical features. Defaults to None.
        weight_on_target (bool, optional): If True will add a class weight based on target_col (see pyspark_ds_toolbox.ml.data_prep.binary_classifier_weights). Defaults to False.
        log_mlflow_run (bool, optional): If True will log params, metrics, confusion matrix, decile table and model in a MLFlow run for each fit. Defaults to False.
        mlflow_experiment_name (Union[None, str], optional): Name of the experiment where the runs should be looged. Defaults to None.
        artifact_stage_path (Union[None, str], optional): Path to write confusion matrix and decile table before logging into mlflow. Defaults to None.

    Raises:
        ValueError: if dfs.schema != dfs_test.schema is True
        ValueError: len(set([id_col, target_col]).difference(set(dfs.columns))) != 0 is True
     
    Returns:
        [dict]: A dict for each algorithm (keys are LogisticRegression, DecisionTreeClassifier, RandomForestClassifier and GBTClassifier). Each element is a dictionary with the keys
            - model: The spark trained model;
            - feature_score: The feature importance of the model (see pyspark.ml.feature_importance.spark_native.extract_features_score); 
            - metrics: dict with confusion_matrix and f1, auc, accuracy, precision, recall and max_ks;
            - decile_table: Table with a decile analysis on the predicted probabilities.
    """
    if dfs.schema != dfs_test.schema:
        raise ValueError('Condition dfs.schema != dfs_test.schema yield True. dfs and dfs_test must same schema.')
    
    if len(set([id_col, target_col]).difference(set(dfs.columns))) != 0:
        raise ValueError('id_col and target_col must be column names in dfs and dfs_test.')
         
    if (log_mlflow_run is True) and (artifact_stage_path is None):
        warnings.warn('log_mlflow_run is True and artifact_stage_path is None. This means that artifacts (confusion matrix and decile table) will not be logged to mlflow.')

    if (mlflow_experiment_name is not None) and (log_mlflow_run is False):
        warnings.warn('mlflow_experiment_name is not None and log_mlflow_run is False. This means that runs will not be logged into the experiment.')

    if (log_mlflow_run is True) and (mlflow_experiment_name is not None):
        mlflow.set_experiment(experiment_name=mlflow_experiment_name)


    print('Computing Features Vector')

    stages = get_features_vector(
        num_features=num_features,
        cat_features=cat_features
    )
    
    if weight_on_target:
        print('Computing Class Weights')
        dfs = binary_classifier_weights(dfs=dfs, col_target=target_col)
        dfs_test = binary_classifier_weights(dfs=dfs_test, col_target=target_col)
        weigth_col = f'weight_{target_col}'

        print('Instanciating Classifiers')
        lr = Pipeline(stages=stages+[spark_cl.LogisticRegression(labelCol=target_col, featuresCol='features', weightCol=weigth_col)])
        dt = Pipeline(stages=stages+[spark_cl.DecisionTreeClassifier(labelCol=target_col, featuresCol='features', weightCol=weigth_col)])
        rf = Pipeline(stages=stages+[spark_cl.RandomForestClassifier(labelCol=target_col, featuresCol='features', weightCol=weigth_col)])
        gbt = Pipeline(stages=stages+[spark_cl.GBTClassifier(labelCol=target_col, featuresCol='features', weightCol=weigth_col)])
        
    else:
        print('Instanciating Classifiers')
        lr = Pipeline(stages=stages+[spark_cl.LogisticRegression(labelCol=target_col, featuresCol='features')])
        dt = Pipeline(stages=stages+[spark_cl.DecisionTreeClassifier(labelCol=target_col, featuresCol='features')])
        rf = Pipeline(stages=stages+[spark_cl.RandomForestClassifier(labelCol=target_col, featuresCol='features')])
        gbt = Pipeline(stages=stages+[spark_cl.GBTClassifier(labelCol=target_col, featuresCol='features')])


    print('Predicting on Test Data and Evaluating')
    out_dict = dict()
    names = [
        'LogisticRegression', 'DecisionTreeClassifier',
        'RandomForestClassifier', 'GBTClassifier'
    ]
    models = [lr, dt, rf, gbt]
    for name, model in tqdm(zip(names, models), total=len(names)):

        model = model.fit(dfs)
        prediction = model.transform(dfs_test)

        df_fi = extract_features_score(model=model.stages[-1], dfs=prediction)
        metrics = cl_eval.binary_classificator_evaluator(
            dfs_prediction=prediction,
            col_target=target_col,
            col_prediction='prediction',
            print_metrics=False
        )

        decile_metrics = cl_eval.binary_classifier_decile_analysis(
            dfs=prediction.withColumn('p1', cl_eval.get_p1(F.col('probability'))),
            col_id=id_col,
            col_target=target_col,
            col_probability='p1'
        )
        decile_metrics = round(decile_metrics.toPandas(), 4)


        out_dict[name] = {
            'model': model,
            'feature_score': df_fi,
            'metrics': metrics,
            'decile_metrics': decile_metrics
        }

        if log_mlflow_run:
            with mlflow.start_run(run_name=f'{name}'):
                # Log Params
                mlflow.log_param('target', target_col)
                mlflow.log_param('weight_on_target', str(weight_on_target))
                paramMap = {str(k).partition('__')[2]: v for (k, v) in model.extractParamMap().items()}
                mlflow.log_params(paramMap)

                # Log Metrics
                mlflow.log_metric("f1", metrics["f1"])
                mlflow.log_metric("auc", metrics["auroc"])
                mlflow.log_metric('accuracy', metrics["accuracy"])
                mlflow.log_metric("precision", metrics["precision"])
                mlflow.log_metric("recall", metrics["recall"])
                mlflow.log_metric("max_ks", float(decile_metrics[['ks']].max()))

                # Log Artefacts and Model
                if artifact_stage_path is not None:
                    metrics['confusion_matrix'].to_csv(f'{artifact_stage_path}confusion_matrix.csv', index=False, sep='|')
                    mlflow.log_artifact(f'{artifact_stage_path}confusion_matrix.csv')

                    decile_metrics.to_csv(f'{artifact_stage_path}decile_metrics.csv', index=False, sep='|')
                    mlflow.log_artifact(f'{artifact_stage_path}decile_metrics.csv')

                    df_fi.to_csv(f'{artifact_stage_path}feature_score.csv', index=False, sep='|')
                    mlflow.log_artifact(f'{artifact_stage_path}feature_score.csv')

                mlflow.spark.log_model(model, "model")

    return out_dict



"""Reshape Tools.

Module dedicated to function to reshape the datasets.
"""

from typeguard import typechecked
import pyspark 
from pyspark.sql import functions as F


@typechecked
def pivot_long(
    dfs: pyspark.sql.dataframe.DataFrame,
    key_column_name: str,
    value_column_name: str,
    key_columns: list,
    value_columns: list,
    print_stack_expr: bool = False
) -> pyspark.sql.dataframe.DataFrame:
    """Function to pivot columns into rows (similar to pandas pivot). See the following link:
    https://sparkbyexamples.com/pyspark/pyspark-pivot-and-unpivot-dataframe/#:~:text=PySpark%20pivot()%20function%20is,individual%20columns%20with%20distinct%20data.

    Args:
        dfs (pyspark.sql.dataframe.DataFrame): A sparkDF.
        key_column_name (str): Name for the column that will receive the columns passed in value_columns as values.
        value_column_name (str): Name the column that will contain the values from the column passed in value_columns.
        key_columns (list): List of columns that will not be 'pivoted'.
        value_columns (list): List os columns to be 'pivoted'
        print_stack_expr (bool, optional): Print the spark sql expression that executes the pivot. Defaults to False.

    Returns:
        [pyspark.sql.dataframe.DataFrame]: The dfs argument with the value_columns 'pivoted'.
    """
    # 1) Creating the stack expression
    len_value_columns = len(value_columns)
    stack_expr = ""
    stack_expr = stack_expr + f'stack({len_value_columns}, '
    for i in range(len_value_columns):
        if i == max(range(len(value_columns))):
            stack_expr = stack_expr + f"'{value_columns[i]}', {value_columns[i]}) as ({key_column_name}, {value_column_name})"
        else:
            stack_expr = stack_expr + f"'{value_columns[i]}', {value_columns[i]}, "
    
    # 2) Printing (or not) the expression
    if print_stack_expr is True:
        print(stack_expr)

    # 3) Creating the pivot    
    dfs_long = dfs.select(key_columns + [F.expr(stack_expr)]).where(f"{value_column_name} is not null")
    return dfs_long

@typechecked
def with_start_week(df: pyspark.sql.dataframe.DataFrame , date_col: str, start_day: str = 'sunday'):
    """Function that adds a date column with the week start.

    Args:
        df (pyspark.sql.dataframe.DataFrame): Spark DataFrame.
        date_col (str): Column name of the column from which the start week date will be computed from.
            Must be one of: pyspark.sql.types.TimestampType, pyspark.sql.types.DateType.
        start_day (str, optional): Which day should the week start?. Defaults to 'sunday'.

    Raises:
        ValueError: If date_col is not from type: pyspark.sql.types.TimestampType.

    Returns:
        [pyspark.sql.dataframe.DataFrame]: The df argument with a column date column called week.
    """
    if not type(df.schema[date_col].dataType) in [pyspark.sql.types.TimestampType, pyspark.sql.types.DateType]:
        raise ValueError(f'Column {date_col} is of wrong type. See the docstring argument.')
    return df.withColumn("week", F.date_sub(F.next_day(F.col(date_col), start_day),7))
    


"""Causal Inference Toolbox

Sub-package dedicated to functionalities related to the field of causal inference. 
For a introduction on the subject see: https://mixtape.scunning.com/introduction.html#what-is-causal-inference
"""

from pyspark_ds_toolbox.causal_inference import ps_matching, diff_in_diff


"""Sub-package Dedicated to Data Preparation for Machine Learning tools.
"""

from pyspark_ds_toolbox.ml.data_prep.class_weights import binary_classifier_weights
from pyspark_ds_toolbox.ml.data_prep.features_vector import get_features_vector



"""This modules implements a way to estimate shap values from SparkDfs.

We are not able to implement a generic function/class for pure spark model, that took in account numerical and categorical features.
So this implementation as a workaround to get the job done. The main function is estimate_shap_values() and you can check directly 
the documentation on this function. 
"""
from typing import Union, List
import pandas as pd
import pyspark
from pyspark.sql import functions as F, types as T, Window as W
import h2o
from h2o.automl import H2OAutoML


class ShapValuesPDF():
    """
    H2O AutoML Wrapper
    """
    def __init__(
        self,
        df: pd.DataFrame,
        id_col: str,
        target_col: str,
        cat_features: Union[None, List[str]],
        sort_metric: str,
        problem_type: str,
        max_mem_size: str = '3G',
        max_models: int = 10,
        max_runtime_secs: int = 60,
        nfolds: int = 5,
        seed: int = 90
    ):
        # Checking for values errors
        if target_col not in df.columns:
            raise ValueError(f'Column {target_col} not in df.columns.')

        if problem_type not in ['classification', 'regression']:
            raise ValueError(f'problem_type: {problem_type}. Should be in ["classification", "regression"].')


        # Params Values
        self.max_mem_size = max_mem_size
        self.df = df
        self.id_col = id_col
        self.target_col = target_col
        self.cat_features = cat_features
        self.sort_metric = sort_metric
        self.problem_type = problem_type
        self.max_models = max_models
        self.max_runtime_secs = max_runtime_secs
        self.nfolds = nfolds
        self.seed = seed
        
        self.df_id = self.df[[self.id_col]]
        # 1) Starting H2O
        self.start_h2o()
        
        # 2) Getting Features Cols
        self.feature_cols = self.get_feature_cols()
        
        # 3) Spliting into Train and Valid
        self.h2o_df = self.as_h2o_df()

        # 4) Training Model
        self.h2o_automl = self.fit_automl()

        # 5) Shap Values
        self.shap_values = self.extract_shap_values()


    def start_h2o(self):
        h2o.init(max_mem_size=self.max_mem_size)

    def get_feature_cols(self):
        return list(set(list(self.df.columns)).difference([self.target_col, self.id_col]))

    def as_h2o_df(self):

        if self.cat_features is not None:
            for c in self.cat_features:
                h2o_df = pd.concat([self.df, pd.get_dummies(self.df[c], drop_first=True, prefix=c)], axis=1)
        else:
            h2o_df = self.df.copy()

        h2o_df = h2o.H2OFrame(h2o_df.drop(columns=[self.id_col]))
        if self.problem_type == 'classification':
            h2o_df[self.target_col] = h2o_df[self.target_col].asfactor()
        
        return h2o_df

    def fit_automl(self):
        ## 1) Training Model
        #https://docs.h2o.ai/h2o/latest-stable/h2o-docs/automl.html
        model = H2OAutoML(
            sort_metric = self.sort_metric,
            max_models = self.max_models, 
            max_runtime_secs = self.max_runtime_secs,
            nfolds = self.nfolds,
            seed = self.seed,
            include_algos = ["GBM", "XGBoost", "DRF"] # models that have shap values.
        )
        model.train(
            x = self.feature_cols,
            y = self.target_col, 
            training_frame = self.h2o_df,
        )

        return model
    
    def extract_shap_values(self):
        fi = self.h2o_automl.leader.predict_contributions(test_data=self.h2o_df).as_data_frame()
        _features = list(fi.columns)
        _features.remove('BiasTerm')

        # Proportional weights
        fi['MeanDeviance'] = fi.drop(columns=['BiasTerm']).sum(axis=1)
        fi = pd.concat([fi[_features].div(fi['MeanDeviance'], axis=0), fi[['MeanDeviance', 'BiasTerm']]], axis=1)
        fi.drop(columns=['MeanDeviance', 'BiasTerm'], inplace=True)
        
        # Wide to Long
        fi = pd.concat([self.df_id.reset_index(drop=True), fi.reset_index(drop=True)], axis=1)
        fi = pd.melt(frame=fi, id_vars=[self.id_col], ignore_index=True, value_name='shap_value')
        return fi

def estimate_shap_values(
    sdf: pyspark.sql.dataframe.DataFrame,
    id_col: str,
    target_col: str,
    cat_features: Union[List[str], None],
    sort_metric: str,
    problem_type: str,
    subset_size: int = 2000,
    max_mem_size: str = '2G',
    max_models: int = 8,
    max_runtime_secs: int = 30,
    nfolds: int = 5,
    seed: int =90
):
    """Computes for each row the shap values of each feature.

    This function will split the sdf into int(sdf.count()/subset_size) pandas dataframes and then use the 
    Class ShapValuesPDF, which is a wrapper of the H2O automl, on each the subseted dataset using the 
    applyInPandas() method of grouped SparkDF.
    
    Check the following link for an intuition of how this works:
    https://www.youtube.com/watch?v=x6dSsbXhyPo 

    Args:
        sdf (pyspark.sql.dataframe.DataFrame): A SparkDF. Note that all column, except the id_col must be features. 
        id_col (str): Column name of the identifier.
        target_col (str): Column name of the target value.
        cat_features (Union[List[str], None]): List of column names of the categorical variables, if any.
        sort_metric (str): A metric to sort the candidates (see https://docs.h2o.ai/h2o/latest-stable/h2o-docs/data-science/algo-params/sort_metric.html).
        problem_type (str): 'regression' or 'classification'. If classification then target_col must be of binary values.
        subset_size (int, optional): Number of rows for each sub dataset. Defaults to 2000.
        max_mem_size (str, optional): Max memory size to be allocated to h2o local cluster. Defaults to '2G'. (see https://docs.h2o.ai/h2o/latest-stable/h2o-py/docs/h2o.html#h2o.init)
        max_models (int, optional): Max number of model to be fitted. These models are ranked according to sort_metric. Defaults to 8. (see https://docs.h2o.ai/h2o/latest-stable/h2o-py/docs/modeling.html?highlight=automl#h2oautoml)
        max_runtime_secs (int, optional): Max number of seconds to spend fitting the models. Defaults to 30. (see https://docs.h2o.ai/h2o/latest-stable/h2o-py/docs/modeling.html?highlight=automl#h2oautoml)
        nfolds (int, optional): Number of folds to be used for cross validation while fitting. Defaults to 5. (see https://docs.h2o.ai/h2o/latest-stable/h2o-py/docs/modeling.html?highlight=automl#h2oautoml)
        seed (int, optional): Seed. Defaults to 90.

    Raises:
        ValueError: if int(sdf.count()/subset_size) < 2 is True

    Returns:
        [pyspark.sql.dataframe.DataFrame]: A sparkDF with the follwing columns
            - id_col: The values from the column passed as id_col;
            - feature: The name of each feature from features_names;
            - shap_value: The shap value of the feature.
    """
    schema = T.StructType([
      T.StructField(id_col, sdf.schema[id_col].dataType),
      T.StructField('variable', T.StringType()),
      T.StructField('shap_value', T.FloatType()),
    ])

    n_quantiles = int(sdf.count()/subset_size)
    if n_quantiles < 2:
        raise ValueError('subset_size must be smaller to result in at least two subsets from sdf. The condition int(sdf.count()/subset_size) < 2 was True')
    sdf = sdf\
        .withColumn('rand', F.rand())\
        .withColumn('qcut', F.ntile(n_quantiles).over(W.partitionBy().orderBy(F.col('rand'))))\
        .withColumn('qcut', F.col('qcut').cast(T.StringType()))

    def __fn(pdf):
        model = ShapValuesPDF(
            df=pdf,
            max_mem_size=max_mem_size,
            id_col=id_col,
            target_col=target_col,
            cat_features=cat_features,
            sort_metric=sort_metric,
            problem_type=problem_type,
            max_models=max_models,
            max_runtime_secs=max_runtime_secs,
            nfolds=nfolds,
            seed=seed
        )
        return model.shap_values

    return sdf.groupBy('qcut').applyInPandas(__fn, schema=schema)



"""Data Quality toolbox.

Module dedicated to provide data quality tools.
"""

from typeguard import typechecked
import pyspark 
from pyspark.sql import functions as F
import pandas as pd


@typechecked
def count_percent_missing_rows_per_column(sdf: pyspark.sql.dataframe.DataFrame) -> pd.core.frame.DataFrame:
  """Computes the percentage of missing values for each column.

  Args:
      dataframe_spark (pyspark.sql.dataframe.DataFrame): A Spark DataFrame.

  Returns:
      pd.core.frame.DataFrame: A Pandas DF where each row representes the sdf columns and the values represents the percentage of null values.
  """
  dataframe_spark_n_rows = sdf.count()
  df_missing_column = sdf\
  .select([F.count(F.when(F.isnan(c) | F.isnull(c), c)).alias(c) for (c,c_type) in sdf.dtypes if c_type not in ('timestamp', 'date')])\
  .toPandas()
  df_missing_column = round(100*(df_missing_column.rename(index={0: 'percent_missing'}).T.sort_values("percent_missing",ascending=False)/dataframe_spark_n_rows),2)  
  return df_missing_column



"""Propensity Score Matching toolbox.

For an introduction on the subject see: https://mixtape.scunning.com/matching-and-subclassification.html#propensity-score-methods
"""
from typeguard import typechecked

import pandas as pd
import pyspark
import pyspark.sql.functions as F
from pyspark.ml import Pipeline
from pyspark.ml.classification import DecisionTreeClassifier, RandomForestClassifier, \
    LogisticRegression, GBTClassifier 

from pyspark_ds_toolbox.ml.classification.eval import get_p1, binary_classifier_decile_analysis


@typechecked
def compute_propensity_score(
    df: pyspark.sql.dataframe.DataFrame,
    treat: str,
    y: str,
    id: str,
    featuresCol: str = 'features',
    train_size: float = 0.8
) -> tuple:
    """Computes the propensity score for a given treatment based on the features from df.

    Args:
        df (pyspark.sql.dataframe.DataFrame): Dataframe with features and treatment assignment.
        treat (str): Column name os the treatment indicating column. Must have value 0 or 1.
        y (str): Column name of the outcome of interest.
        id (str): Id of obervations.
        featuresCol (str, optional): Assembled column to be used in a pyspark pipelie. Defaults to 'features'.
        train_size (float, optional): Proportion of the dataset to be used in training. Defaults to 0.8.

    Returns:
        tuple: 1 a sparkDF with the id, propensity score, treat and y columns;
               2 a pandasDF with the evaluation of the models used to compute the propensity score. 
    """
    train, test = df.select(id, featuresCol, treat, y).randomSplit([train_size, (1-train_size)], seed=12345)

    spark_classifiers = {
        'logistic_regression': LogisticRegression(featuresCol=featuresCol, labelCol=treat),
        'decision_tree': DecisionTreeClassifier(featuresCol=featuresCol, labelCol=treat),
        'random_forest': RandomForestClassifier(featuresCol=featuresCol, labelCol=treat),
        'gradient_boosting': GBTClassifier(featuresCol=featuresCol, labelCol=treat)
    }
    df_evaluate = pd.DataFrame()

    for classifier_name, classifier in spark_classifiers.items():
        print(f'{classifier_name}: Starting')
        pipeline = Pipeline(stages = [classifier])
        
        # Fit no Modelo e Predict no test
        print(f'{classifier_name}: Fitting Pipeline')
        fitted_classifier = pipeline.fit(train)
        print(f'{classifier_name}: Making Predictions on test data')
        prediction_on_test = fitted_classifier.transform(test)\
            .withColumn('ps', get_p1(F.col('probability')))

        # Metricas e Avaliacao
        eval_decile = binary_classifier_decile_analysis(dfs=prediction_on_test, col_id='index', col_target=treat, col_probability='ps').toPandas()
        max_ks = eval_decile.iloc[eval_decile['ks'].idxmax(),]
        df_temp = pd.DataFrame({
            'model':[classifier_name],
            'ks_max': [max_ks['ks']],
            'at_decile': [max_ks['percentile']],
            'precision': [max_ks['precision_at_percentile']],
            'recall': [max_ks['cum_eventrate']]
        })
        df_evaluate = df_evaluate.append(df_temp)
        
    final_model = spark_classifiers[df_evaluate.iloc[df_evaluate['ks_max'].idxmax()]['model']].fit(df)
    df_final = final_model.transform(df).withColumn('ps', get_p1(F.col('probability')))
    df_final = df_final.select(id, 'ps', treat, y)
    return (df_final, df_evaluate)

@typechecked
def estimate_causal_effect(
    df_ps: pyspark.sql.dataframe.DataFrame,
    y: str,
    treat: str,
    ps: str
) -> float:
    """Function that estimates the ATE based on propensity scores.

    The implementation is based on chapter 5 Matching and Subclassification
    section 5.3 Approximate Matching of the book Causal Inference: the Mixtape
    (https://mixtape.scunning.com/index.html).

    Args:
        df_ps (pyspark.sql.dataframe.DataFrame): SparkDF with outcome of interest, treatment and propensity scores.
        y (str): Column name of the outcome of interest.
        treat (str): Column name indicating receivement of treatment or not.
        ps (str): Column name with propensity score.

    Returns:
        float: The ATE.
    """
    N = df_ps.count()

    df_effect = df_ps\
        .filter(f'{ps} > 0.1 and {ps} < 0.9')\
        .withColumn('d1', F.expr(f'{treat}/{ps}'))\
        .withColumn('d0', F.expr(f'(1-{treat})/(1-{ps})'))

    N = df_effect.count()
    s1 = df_effect.select(F.sum('d1')).collect()[0][0]
    s0 = df_effect.select(F.sum('d0')).collect()[0][0]

    df_effect = df_effect\
        .withColumn('y1', (F.col(treat)*F.col(y)/F.col(ps)) / (s1/N))\
        .withColumn('y0', ((1-F.col(treat))*F.col(y))/(1-F.col(ps)) / (s0/N))\
        .withColumn('ht', F.expr('y1 -y0'))\
        .drop('y1', 'y0', 'd1', 'd0')
    
    return df_effect.select(F.avg('ht')).collect()[0][0]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyspark_ds_toolbox',
 'pyspark_ds_toolbox.causal_inference',
 'pyspark_ds_toolbox.ml',
 'pyspark_ds_toolbox.ml.classification',
 'pyspark_ds_toolbox.ml.data_prep',
 'pyspark_ds_toolbox.ml.feature_importance',
 'pyspark_ds_toolbox.stats',
 'pyspark_ds_toolbox.wrangling']

package_data = \
{'': ['*']}

install_requires = \
['h2o>=3.34.0,<4.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'mlflow>=1.22.0,<2.0.0',
 'numpy==1.21.0',
 'pandas>=1.3.4,<2.0.0',
 'pyarrow>=6.0.1,<7.0.0',
 'pyspark>=3.2',
 'seaborn>=0.11.2,<0.12.0',
 'tqdm>=4.62.3,<5.0.0',
 'typeguard>=2.13.2,<3.0.0']

setup_kwargs = {
    'name': 'pyspark-ds-toolbox',
    'version': '0.3.4',
    'description': 'A Pyspark companion for data science tasks.',
    'long_description': '# Pyspark DS Toolbox\n\n<!-- badges: start -->\n[![Lifecycle:\nexperimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)\n[![PyPI Latest Release](https://img.shields.io/pypi/v/pyspark-ds-toolbox.svg)](https://pypi.org/project/pyspark-ds-toolbox/)\n[![CodeFactor](https://www.codefactor.io/repository/github/viniciusmsousa/pyspark-ds-toolbox/badge)](https://www.codefactor.io/repository/github/viniciusmsousa/pyspark-ds-toolbox)\n[![Maintainability](https://api.codeclimate.com/v1/badges/9a85a662305167c5aba1/maintainability)](https://codeclimate.com/github/viniciusmsousa/pyspark-ds-toolbox/maintainability)\n[![Codecov test coverage](https://codecov.io/gh/viniciusmsousa/pyspark-ds-toolbox/branch/main/graph/badge.svg)](https://codecov.io/gh/viniciusmsousa/pyspark-ds-toolbox?branch=main)\n[![Package Tests](https://github.com/viniciusmsousa/pyspark-ds-toolbox/actions/workflows/package-tests.yml/badge.svg)](https://github.com/viniciusmsousa/pyspark-ds-toolbox/actions)\n[![Downloads](https://pepy.tech/badge/pyspark-ds-toolbox)](https://pepy.tech/project/pyspark-ds-toolbox)\n<!-- badges: end -->\n\n\nThe objective of the package is to provide a set of tools that helps the daily work of data science with spark. The documentation can be found [here](https://viniciusmsousa.github.io/pyspark-ds-toolbox/index.html). Feel free to contribute :)\n\n\n## Installation\n\nDirectly from PyPi:\n```\npip install pyspark-ds-toolbox\n```\n\nor from github, note that installing from github will install the latest development version:\n```\npip install git+https://github.com/viniciusmsousa/pyspark-ds-toolbox.git\n```\n\n## Organization\n\nThe package is currently organized in a structure based on the nature of the task, such as data wrangling, model/prediction evaluation, and so on.\n\n```\npyspark_ds_toolbox         # Main Package\n├─ causal_inference           # Sub-package dedicated to Causal Inferece\n│  ├─ diff_in_diff.py   \n│  └─ ps_matching.py    \n├─ ml                         # Sub-package dedicated to ML\n│  ├─ data_prep                  # Sub-package to ML data preparation tools\n│  │  ├─ class_weights.py     \n│  │  └─ features_vector.py \n│  ├─ classification             # Sub-package decidated to classification tasks\n│  │  ├─ eval.py\n│  │  └─ baseline_classifiers.py \n│  └─ feature_importance         # Sub-package with feature importance tools\n│     ├─ native_spark.py\n│     └─ shap_values.py    \n├─ wrangling                  # Sub-package decidated to data wrangling tasks\n│  ├─ reshape.py               \n│  └─ data_quality.py         \n└─ stats                      # Sub-package dedicated to basic statistic functionalities\n   └─ association.py    \n```\n\n',
    'author': 'vinicius.sousa',
    'author_email': 'vinisousa04@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/viniciusmsousa/pyspark-ds-toolbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)
